#ifdef	KANJI
#include <stdio.h>
#include <stdlib.h>

#if	defined(__TURBOC__) || defined(__MSC__)
#include	<jctype.h>
#include	<jstring.h>
#endif

#include "psort.h"

#ifndef	__JSTRING_H
   #ifdef EUC
   #define UC(c)        ((unsigned char)(c))
   #define iskanji(c)   ((c) & 0x80)
   #else /* sjis */
   #define UC(c)        ((unsigned char)(c))
   #define BD(l, h, c)  (((l) <= (c))&&((c) <= (h)))
   #define iskanji(c)   (BD(0x81, 0x9F, UC(c)) || BD(0xE0, 0xFC, UC(c)))
   #endif
#endif

int kanji_in;
int kanji_out;
int kana8;

uchar esc_kanji_in[]  = "\x1b$B";
uchar esc_kanji_out[] = "\x1b(J";

void KmodeRead( void )
{
	f_jis &= ~0xf;
	f_jis |= ((f_jis & 0xf00) >> 8);
	kana8 &= ~(J_KANA8|J_KANAQ);
}

void JISmodeWrite( void )
{
	f_jis |= JO_JIS;
	if(kanji_in)
		esc_kanji_in[ESC_POS] = kanji_in;
	if(kanji_out)
		esc_kanji_out[ESC_POS] = kanji_out;
	if(kana8 & J_KANA8)
		kana8 |= JO_KANA8;
}

#ifdef	EUC
static uchar *
get_room(uchar *s, uchar *e)
{
	int	len, dif;

	len = strlen(s) + 1;
	dif = (e - s) - len;
	if(dif <= 0)
		return NULL;
	memmove(s + dif, s, len);
	return(s + dif);
}
#endif

/*
 * 		JIS/EUC -> SHIFT JIS
 */
uint	jtosj( uint code )
{
	uint	code_l;

	code_l = code & 0x7f;

	if(code & 0x0100){
		if(code_l >= 0x60) code_l++;
		code = ((code >> 1) & 0x3f00) + code_l + 0x711f;
	}
	else	code = ((code >> 1) & 0x3f00) + code_l + 0x707e;
	return((code < 0xa000)? code : (code + 0x4000));
}

#define		DEL		0x7f
#define		SP		0x20
#define		LF		'\n'
#define		CR		0x0d

/*
 * 		SHIFT JIS -> EUC
 */
uint	sjtoeuc( uint code )
{
	uint	code_h;

	code_h = code & 0xff00;
	code += code_h - ((code_h <= 0x9fff)?0x6100:0xe100);

	if((code & 0x00ff) > 0x9f)
		return (code + 0x102);
	return (((code & 0x00ff) > 0x7f)? (code + 0x60):(code + 0x61));
}

int 
#ifdef	EUC
to_euc(uchar *s, int size)
#else
to_euc(uchar *s)
#endif
{
	uchar *d, *top;
	int	ch, flag, ch2, ch3, k_flag;
#ifdef	EUC
	uchar *e;

start:
	e = (d = top = s) + size;
#else
start:
	d = top = s;
#endif

	k_flag = 0;

#ifdef	SJIS
	if(f_jis & J_SJIS)						/* ShiftJIS */
		return 0;
	if((f_jis & J_TYPE) == J_EUC){			/* EUC -> ShiftJIS */
		while((ch = *s++) != 0){
			if(!(ch & 0x80)){				/* normal */
				*d++ = ch;
				continue;
			}
			if(!((ch2 = *s++) & 0x80))
				return 1;
			if(ch == EUC_KANA)				/* kana */
				*d++ = ch2;
			else{							/* kanji */
				ch = jtosj((ch<<8)|ch2);
				*d++ = (ch>>8);
				*d++ = (ch & 0xff);
			}
		}
		goto end;
	}
#else
	if((f_jis & J_TYPE) == EUC)				/* EUC -> EUC */
		return 0;

	if(f_jis & J_SJIS){
		while((ch = *s++) != 0){			/* ShiftJIS -> EUC */
			if(!(ch & 0x80)){
				if(ch != CR || !(f_jis & J_CR))
					*d++ = ch;
				continue;
			}
			if((ch >= 0x81 && ch <= 0x9f) || (ch >= 0xe0 && ch <= 0xfc)){
				ch2 = *s++;
				if(!ch2) return 1;
				ch = sjtoeuc((ch<<8)|ch2);
				*d++ = (ch>>8);
				*d++ = ch & 0xff;
			}else{
				*d++ = EUC_KANA;
				if(d == s && (s = get_room(s, e)) == NULL)
				return 2;
				*d++ = ch;
			}
		}
		goto end;
	}
#endif

eng:
	while((ch = *s++) != 0){
		if(!(ch & 0x80)){					/*** normal ***/
			if(ch != ESC){
				if(ch != KANA_IN){
					if(ch != CR || !(f_jis & J_CR))
eng1:					*d++ = ch;
					continue;
				}
				goto kana;					/* -> kana7 */
			}
			if((ch = *s++) != '$'){			/* ESC */
eng2:			*d++ = ESC;
				goto eng1;
			}
			if((ch = *s++) == 'B' || ch == '@' || ch == esc_kanji_in[ESC_POS]){
				if(!kanji_in){
					kanji_in = ch;
					f_jis |= J_JIS;
					if(kana8 & J_KANAQ) return 3;
				}
				goto kanji;					/* -> kanji */
			}
			*s++ = ch;
			goto eng2;
		}
#ifdef	EUC
		k_flag |= (J_KANAQ|J_KANA8);			/*** kana8 or EUC (or SJIS) ***/
		if(f_jis & 1){
			*d++ = EUC_KANA;
			if(d == s && (s = get_room(s, e)) == NULL)
			return 2;
		}
		*d++ = ch;
		if(!kana8 && (f_jis & J_Q7)){
			if(ch >= 0xa1 && ch != EUC_KANA && *s >= 0xa1)
				*d++ = *s++;				/* treated as EUC */
			else if((ch >= 0x81 && ch <= 0x9f) || (ch >= 0xe0 && ch <= 0xfc)){
				f_jis |= J_SJIS;
				s = top;					/* Assume to be SJIS */
				goto start;
			}
		}
#else									/*** kana8 or SJIS (or EUC) */
		*d++ = ch;
		if(iskanji(ch) && *s){
			k_flag |= J_KANAQ;
			*d++ = *s++;				/* treated as SJIS if ? */
		}else{
			if(!kana8 && (f_jis & J_Q7) && ch >= 0xa1 && *s >= 0xa1){
				f_jis &= ~J_TYPE;		/* treated as EUC */
				s = top;
				goto start;
			}
			k_flag |= J_KANA8;
		}
#endif
	}
end:
	kana8 |= k_flag;
	*d = 0;
	return 0;

kanji:										/*** kanji ***/
	while((ch = *s++) != 0){
		ch2 = *s++;
		if(ch <= 0x20){
			if(ch == ESC && ch2 == '('){
				if((ch3 = *s++) == 'J' || ch3 == 'B' || ch3 == 'H' ||
					ch3 == esc_kanji_out[ESC_POS]){
					if(!kanji_out)
						kanji_out = ch;
					goto eng;				/* -> normal */
				}
			}
			return 1;
		}else{
			if(!ch2 || ((ch|ch2) & 0x80))
				return 1;
#ifdef SJIS
			ch = jtosj((ch<<8)|ch2);
			*d++ = (ch>>8);
			*d++ = ch & 0xff;
#else
			*d++ = ch  | 0x80;
			*d++ = ch2 | 0x80;
#endif
		}
	}
	goto end;

kana:										/*** kana7 ***/
	if(kana8 & (J_KANA8|J_KANAQ))	return 3;
	f_jis |= 1;
	while((ch = *s++) != 0){
		if(ch <= 0x20){
			if(ch == KANA_OUT)	goto eng;	/* -> normal */
			*d++ = ch;
			continue;
		}
#ifdef	EUC
		*d++ = EUC_KANA;
		if(d == s && (s = get_room(s, e)) == NULL)
			return 2;
#endif
		*d++ = ch | 0x80;
	}
	goto end;
}


int
to_jis(uchar *s)
{
	int ch, ch2, flag, result;

	flag = result = 0;

#ifdef	SJIS
	if(f_jis & JO_SJIS){					/* ShiftJIS -> ShiftJIS */
sjis:	if(!(f_jis & JO_CR))
			return(fputs(s, fp_out));
		while((ch=*s++) != 0){
			if(ch == LF)
				putc(CR, fp_out);
			result |= putc(ch, fp_out);
		}
		return result;
	}
	if((f_jis & JO_TYPE) == JO_EUC){
euc:	while((ch = *s++) != 0){			/* ShiftJIS -> EUC */
			if(!(ch & 0x80)){
				result |= putc(ch, fp_out);
				continue;
			}
			if(iskanji(ch)){
				ch2 = *s++;
				if(!ch2) return 1;
				ch = sjtoeuc((ch<<8)|ch2);
				result |= putc(ch >> 8, fp_out);
				result |= putc(ch & 0xff, fp_out);
			}else{
				result |= putc(EUC_KANA, fp_out);
				result |= putc(ch, fp_out);
			}
		}
		return result;
	}
	if((f_jis & JO_TYPE) == JO_Q){
		if(!(f_jis & J_TYPE)){
			f_jis &= ~(J_TYPE << 4);
			goto euc;
		}
		if(!(f_jis & J_JIS)) goto sjis;
		f_jis |= JO_JIS;
		if(kanji_in)
			esc_kanji_in[ESC_POS] = kanji_in;
		if(kanji_out)
			esc_kanji_out[ESC_POS] = kanji_out;
		if(kana8 & J_KANA8)
			kana8 |= JO_KANA8;
	}
#else
	if((f_jis & JO_TYPE) == JO_EUC)
euc:		return(fputs(s, fp_out));				/* EUC -> EUC */

	if(f_jis & JO_SJIS){							/* EUC -> ShiftJIS */
sjis:	while((ch = *s++) != 0){
			if(ch & 0x80){
				if(!(ch2 = *s++))
					return result;
				if(ch == EUC_KANA)
					putc(ch2, fp_out);
				else{
					ch = jtosj((ch<<8)|ch2);
					putc(ch>>8, fp_out);
					putc(ch & 0xff, fp_out);
				}
			}else{
				if(ch == LF && (f_jis & 0x80))
					putc(CR, fp_out);
				result |= putc(ch, fp_out);
			}
		}
		return result;
	}
	if((f_jis & JO_TYPE) == JO_Q){
		if(f_jis & J_SJIS){
			f_jis |= JO_SJIS;
			goto sjis;
		}
		if(!(f_jis & J_JIS)) goto euc;
		JISmodeWrite();
	}
#endif

	while((ch = *s++) != 0){						/* EUC/SJIS -> JIS */
		if(!(ch & 0x80)){							/*** NORMAL ***/
			if(flag){
				if(flag == 1){						/* <- KANJI */
					result |= fputs( esc_kanji_out, fp_out);
				}else								/* <- !NORMAL */
					putc( KANA_OUT, fp_out );
				flag = 0;
			}
			result |= putc(ch, fp_out);
			continue;
		}else{
			if(!*s) break;
#ifdef	EUC
			if(ch == EUC_KANA)						/*** KANA ***/
#else
			if(!iskanji(ch))
#endif
			{
				if(kana8 & JO_KANA8){				/** kana8 **/
					if(flag == 1){					/* <- KANJI */
						flag = 0;
						result |= fputs( esc_kanji_out, fp_out );
					}
#ifdef	EUC
					putc(*s++, fp_out);
#else
					putc(ch, fp_out);
#endif
					continue;
				}else{								/** kana7 **/
					if(flag == 1)					/* <- KANJI */
						result |= fputs( esc_kanji_out, fp_out );
					if(flag != 2){					/* <- NOMAL or KANJI */
						flag = 2;
						putc(KANA_IN, fp_out);
					}
					goto mask;
				}
			}
													/*** KANJI ***/
			if(flag == 2)
				putc( KANA_OUT, fp_out );			/* <- KANA */
			if(flag != 1){							/* <- !KANJI */
				flag = 1;
				fputs( esc_kanji_in, fp_out );
			}
#ifdef	EUC
			putc(ch & 0x7f, fp_out);
mask:		result |= putc(*s++ & 0x7f, fp_out);
#else
			ch = sjtoeuc((ch << 8)|*s++);
			putc((ch>>8) & 0x7f, fp_out);
mask:		putc(ch & 0x7f, fp_out);
#endif
		}
	}
	if(flag == 2)
		result |= putc( KANA_OUT, fp_out );			/* <- KANA */
	else if(flag == 1)
		result |= fputs( esc_kanji_out , fp_out );	/* <- KANJI */
	return result;
}

#endif
